:- use_module(library(time)).

:-dynamic(amFacutBonus/0).
:-dynamic(parse/2).
:-dynamic(describe/1).
:-dynamic(solve/2).
:-dynamic(longform/0).

:-ensure_loaded('transmission.pl').


ttimeLimit(5). % in seconds, per test

longform. % keep this for verbose test output.

% Test using:
% ttestAll -- runs all tests
% ttestOne(TestID) -- runs 1 test
% ttestOneA(TestID) -- runs 1 test, with no time limit


% ====== points
% points(+For, -Points)
ttpoints(parse, Points, OtherPoints) :- Points is OtherPoints * 20 / 75.
ttpoints(describe, "+0.0: Points for describe are given upon presentation.~n").
% points(+TestID, -Points)
ttpoints(11, 1.0).
ttpoints(12, 1.3).
ttpoints(T2X, 0.5) :- member(T2X, [21, 22, 23, 24]).
ttpoints(T3X, 0.8) :- member(T3X, [31, 33, 34, 37]).
ttpoints(T4X, 0.4) :- member(T4X, [41, 42, 43, 45, 411]).


% ====== tests
% tt(+TestID, -Raw)

% a+  b-
tt(11, [sources([a]), receivers([b])]).

%  a+  b-
%  c-  d+
tt(12, [sources([a, d]), receivers([b, c])]).

% ------------------------
%       a+
%  b=
%       c-
tt(21, [sources([a]), transceivers([b]), receivers([c])]).

%         a-
%  b+  c=   d+   e=
%         f-
tt(22, [sources([b, d]), receivers([a, f]), transceivers([c, e]),
	collinear([[b, c, d, e]])]).

%   a+       b+
%       c==
%   d-       e-
tt(23, [sources([a, b]), receivers([d, e]), transceivers([c, 2]),
	collinear([[a, c, e], [b, c, d]])]).

%         a+
%  b-  c=   d==   e-
%         f+
tt(24, [sources([a, f]), receivers([b, e]), transceivers([c, d, 2]),
	collinear([[b, c, d, e]])]).

% ------------------------
%       a++++
%  b--  c==
%  d=   e--    f=
tt(31, [sources([a, 4]), receivers([b, 2, e, 2]), transceivers([c, 2, d, f]),
        collinear([[a, c, e], [d, e, f]])]).

%      a==
%  b==     d==
%      c=+
tt(33, [transceivers([a, 2, b, 2, c, 1, d, 2]), source_transceivers([c])]).

%    a=    b=+
%    c=+   d=-
tt(34, [transceivers([a, b, c, d]), source_transceivers([b, c]), star_transceivers([d])]).

%       a==
%           b=
%  c====        d====-
%       e-
%
%  f=+  g=+
tt(37, [transceivers([a, 2, b, c, 4, d, 4, f, g]), source_transceivers([f, g]),
	star_transceivers([d]), receivers([e]), collinear([[a, b, d], [a, e, g], [f, e, b]])]).

% bonus
% a-   b-
%   c#   d-
%
%   e+
tt(41, [sources([e]), antennas([[c, a, b, d]]), receivers([a, b, d])]).

% a-  b#  c--  d#  e-
%
%
%         f++
%      g=     h=
tt(42, [sources([f, 2]), antennas([[b, a, c], [d, c, e]]), receivers([a, c, 2, e]),
	transceivers([g, h]), collinear([[a, b, c, d, e]])]).

%      a--       b==
%  c-- d#        e#  f=
%           g+
tt(43, [antennas([[d, a, c], [e, f, b]]), sources([g]), receivers([a, 2, c, 2]),
	transceivers([b, 2, f]), collinear([[c, d, e, f]])]).

%   a=  b=+
%
%   c-- d# e==   f=-
tt(45, [antennas([[d, e, c]]), receivers([c, 2]), transceivers([a, b, e, 2, f]),
       source_transceivers([b]), star_transceivers([f]), collinear([[c, d, e, f]])]).

%     a++   b=
%                         c=
%
%   d*  e==== f*  g----  h*
%  i---                     j=
%               k+
tt(411, [sources([a, 2, k]), receivers([g, 4, i, 3]),
	transceivers([b, c, e, 4, j]),
	antennas([[d, e, i], [f, e, g], [h, c, j, g]]),
	collinear([[d, e, f, g, h]])]).






% Internal test data
% TestID, all nodes, sources, transceivers, receivers, unavailable edges
ttc(11, [a, b], [a], [], [b], []).
ttc(12, [a, b, c, d], [a, d], [], [b, c], []).
ttc(21, [a, b, c], [a], [b], [c], []).
ttc(22, [a, b, c, d, e, f], [b, d], [c, e], [a, f], [[b, d], [c, e], [b, e]]).
ttc(23, [a, b, c, d, e], [a, b], [c, c], [d, e], [[b, d], [a, e]]).
ttc(24, [a, b, c, d, e, f], [a, f], [c, d, d], [b, e], [[b, d], [c, e], [b, e]]).
ttc(31, [a, b, c, d, e, f], [a, a, a, a], [c, c, d, f], [b, b, e, e], [[a, e], [d, f]]).
ttc(33, [a, b, c, d], [c], [a, a, b, b, c, d, d], [], []).
ttc(34, [a, b, c, d], [b, c], [a, b, c, d], [d], []).
ttc(37, [a, b, c, d, e, f, g], [f, g],
    [a, a, b, c, c, c, c, d, d, d, d, f, g], [e, d], [[a, d], [f, b], [a, g]]).
% same + antennas (as in normal test)
ttcb(41, [a, b, c, d, e], [e], [], [a, b, d], [], [[c, a, b, d]]).
ttcb(42, [a, b, c, d, e, f, g, h], [f, f], [g, h], [a, c, c, e],
     [[a, c], [a, d], [a, e], [b, d], [b, e], [c, e]], [[b, a, c], [d, c, e]]).
ttcb(43, [a, b, c, d, e, f, g], [g], [b, b, f], [a, a, c, c], [[c, e], [c, f], [d, f]],
    [[d, a, c], [e, f, b]]).
ttcb(45, [a, b, c, d, e, f], [b], [a, b, e, e, f], [c, c, f], [[c, e], [c, f], [d, f]],
     [[d, e, c]]).
ttcb(411, [a, b, c, d, e, f, g, h, i, j, k], [a, a, k], [b, c, e, e, e, e, j],
    [i, i, i, g, g, g, g], [[d, f], [d, g], [d, h], [e, g], [e, h], [f, h]],
    [[d, e, i], [f, e, g], [h, c, j, g]]).


ttPrintIfBonus(Tense) :- (   amFacutBonus ->
	format("+0.0: ~w WITH BONUS (amFacutBonus is true).~n", [Tense])
	;
	format("+0.0: ~w WITHOUT BONUS (amFacutBonus is false).~n", [Tense])
	).

% Runs all tests.
ttestAll :-
	ttPrintIfBonus('Testing'),
	findall(ID, (
		    tt(ID, _),
		    (	amFacutBonus -> true ; ID < 40 )
		), AllTests),
	length(AllTests, LT),
	findall(IDD, (member(IDD, AllTests), ttestOne(IDD)), L),
	length(L, LS), findall(T, (member(T, AllTests), \+ member(T, L)), LF),
	format("+0.0: Result: ~w / ~w tests succeeded: ~w. tests failed: ~w.~n", [LS, LT, L, LF]),
	findall(Points, (member(IDD, L), ttpoints(IDD, Points)), LPoints),
	sum_list(LPoints, PointsTests),
	findall(Points, (member(IDD, L), IDD < 40, ttpoints(IDD, Points)), LPointsParse),
	sum_list(LPointsParse, PointsTestsParse),
	ttpoints(parse, PointsParse, PointsTestsParse), ttpoints(describe, PointsDescr),
	format("+0.0: Points for parse (computed from non-bonus tests): ~w~n", [PointsParse]),
	format(PointsDescr),
	format("+0.0: Points for tests: ~w~n", [PointsTests]),
	TPoints is PointsTests + PointsParse,
	format("+0.0: Total points (w/o describe): ~w.~n", [TPoints]),
	ttPrintIfBonus('Tested'),
	!.

% ttestOne(TestID) -- runs 1 test
ttestOne(TestID) :- ttimeLimit(T), catch(
				       call_with_time_limit(T, ttestOneA(TestID)),
				       time_limit_exceeded,
					 (
					   format("+0.0: Time limit [~w] exceeded.~n", [T]),
					   !, fail
				       )).

sFormat(Mesg, L) :- longform, !, format(Mesg, L).
sFormat(_, _).
sFormat(Mesg) :- sFormat(Mesg, []).

pre :- format("+0.0: ").

% ttestOneA(TestID) -- runs 1 test, with no time limit
ttestOneA(TestID) :- \+ tt(TestID, _), format("Test ~w not found.~n", [TestID]).
ttestOneA(TestID) :- tt(TestID, Raw),
	sFormat("~n~n===========================~nTest ~w: ", [TestID]),
	ttestParse(Raw, Problem),
	ttestDescribe(Problem),
	ttestSolve(Problem, Solution),
	ttCheckSolution(TestID, Solution),
	ttpoints(TestID, Points),
	format("+~w: Test ~w OK.~n", [Points, TestID]).
ttestOneA(TestID) :- !, format("+0.0: Test ~w failed.~n", [TestID]), fail.

ttestParse(Raw, Problem) :- parse(Raw, Problem), !, sFormat(".").
ttestParse(_, _) :- pre, format("parse predicate failed~n"), !, fail.

ttestDescribe(Problem) :- sFormat("<~n"), describe(Problem), !, sFormat("~n>.").
ttestDescribe( _) :- pre, format("describe predicate failed~n"), !, fail.

ttestSolve(Problem, Solution) :- solve(Problem, Solution), !, sFormat(".").
ttestSolve(_, _) :- pre, format("solve predicate failed~n"), !, fail.


% ttCheckSolution(+TestID, +Solution)
ttCheckSolution(TestID, Edges) :-
	(   ttc(TestID, NN, SS, TT, RR, XE), BCast = []; ttcb(TestID, NN, SS, TT, RR, XE, BCast);
	pre, format("internal error checker test ~w not found.~n", [TestID])),
	sFormat("given solution: ~n ~w ~nchecking", [Edges]),
	(   checkEE(Edges, [NN, SS, TT, RR, XE, BCast]), !, sFormat(".");
	pre, format("Edge check 1 failed.~n"), !, fail),
	(   checkE2(Edges), !, sFormat("."); pre, format("Edge check 2 failed.~n"), !, fail),
	(   checkGraph(SS, Edges, BCast), !, sFormat(".");
	pre, format("Graph coherency check failed.~n"), !, fail),
	(   amFacutBonus -> fillAntennas(BCast, TT, RR, Edges, Edges2),
	    sFormat("w/ edges from antennas: ~w~n", [Edges2]),
	    findall(N, (member(N, NN), \+ member([N|_], BCast)), NN2)
	;
	    Edges2 = Edges, NN2 = NN
	),
	(   checkNodes(Edges2, NN2, SS, TT, RR), !, sFormat(".");
	pre, format("Node check failed.~n"), !, fail ),
	sFormat("OK.~n")
	.

% checks that edges go between the right kinds of nodes
checkEE([], _).
checkEE([E|EE], A) :- checkE(E, A), !, checkEE(EE, A).

% edge nodes are actually nodes
checkE(E, [NN, _, _, _, _, _]) :- [X, _, _] = E, \+ member(X, NN), e1(X, E), !, fail.
checkE(E, [NN, _, _,_,  _, _]) :- [_, Y, _] = E, \+ member(Y, NN), e1(Y, E), !, fail.
% edges have strictily positive width
checkE(E, _) :- [_, _, W] = E, W =< 0, e2(W, E), !, fail.
% edges are not forbidden (going through other nodes)
checkE(E, [_, _, _, _, XE, _]) :- [X, Y, _] = E, (member([X, Y], XE) ; member([Y, X], XE)), e3(E), !, fail.
% edges don't go out of receivers (that are not actually transceivers)
checkE(E, [_, _, TT, RR, _, _]) :- [X, _, _] = E, member(X, RR), \+ member(X, TT), e6a(E), !, fail.
% edges don't go out of broadcast antennas
checkE(E, [_, _, _,  _, _, BCast]) :- [X, _, _] = E, member(X, BCast), e6b(E), !, fail.
% edges don't go into sources (that are not actually transceivers)
checkE(E, [_, SS, TT, _, _, _]) :- [_, Y, _] = E, member(Y, SS), \+ member(Y, TT), e6c(E), !, fail.

checkE([X, Y, _], [NN, SS, TT, RR, XE, BCast]) :- member(X, NN), member(Y, NN),
	(   \+ member(X, RR); member(X, TT)),
	\+ member(X, BCast), (\+ member(Y, SS); member(Y, TT)),
	\+ member([X, Y], XE), \+ member([Y, X], XE).

% checks that edges don't appear multiple times
checkE2([]):-!.
checkE2([E|EE]):-[X, Y, _] = E, member([X, Y, _], EE), !, e4(E), fail.
checkE2([E|EE]):-[X, Y, _] = E, member([Y, X, _], EE), !, e5(E), fail.
checkE2([[X, Y, _]|EE]) :- \+ member([X, Y, _], EE), \+ member([Y, X, _], EE),
	checkE2(EE).

% adds all edges from antennas to nodes in range
fillAntennas([], _, _, EE, EE).
fillAntennas([[A|Array]|BB], TT, RR, EE, EEO2) :-
	findall(WI, member([_, A, WI], EE), LWI), sum_list(LWI, WTI),
	findall([A, X, WO], (
		    member(X, Array),
		    \+ member([X, A, _], EE),
		    findall(X, member(X, TT), LT), length(LT, WTT),
		    findall(X, member(X, RR), LR), length(LR, WRR),
		    findall(W, member([_, X, W], EE), LXA), sum_list(LXA, WXA),
		    WO is min(WTI, WTT + WRR - WXA)
		), EEA), append(EE, EEA, EEO),
	fillAntennas(BB, TT, RR, EEO, EEO2).

% builds paths starting from sources and going through edges in the
% solution, with the target of exhausting all provided edges.
checkGraph(SS, EE, BB) :- checkGraph(SS, EE, [], R, BB),
	(  R == [];
	format("Can't include these edges in paths starting from sources: ~w~n", [R]), ! , fail ).
checkGraph([S | Sources], [[S, X, W] | PEdges], REdges, R, BB) :- dec(W, W1),
	append(PEdges, REdges, Edges),
	(   W > 1, Edges2 = [[S, X, W1] | Edges]; Edges2 = Edges),
	checkGraph(X, Edges2, [], R1, BB),
	checkGraph(Sources, R1, [], R, BB).
checkGraph(X, Edges, [], R, BB) :-
	member([X|Array], BB),
	checkGraph(Array, Edges, [], R, BB).
checkGraph(X, [[X, Y, W] | PEdges], REdges, R, BB) :- dec(W, W1),
	append(PEdges, REdges, Edges),
	(   W > 1, Edges2 = [[X, Y, W1] | Edges]; Edges2 = Edges),
	checkGraph(Y, Edges2, [], R, BB).
checkGraph(X, [E | PEdges], REdges, R, BB) :- checkGraph(X, PEdges, [E | REdges], R, BB).
checkGraph(_, [], Edges, Edges, _).
dec(X, X1):-X1 is X - 1.


checkNodes(_, [], _, _, _) :- !.
% checks that the sources can provided tha information extracted from
% them
checkNodes(EE, [N|NN], SS, TT, RR) :-
	member(N, SS), \+ member(N, TT), \+ member(N, RR), !,
	findall(W, (member(E, EE), E = [N, _, W]), LW),
	findall(N, member(N, SS), LS), length(LS, WS),
	sum_list(LW, WT),
	(    WT =< WS; eN1(N, WT, WS), !, fail),
	checkNodes(EE, NN, SS, TT, RR).
% checks that the transceivers have incoming width equal to their width,
% and outgoing width lower or equal to their width.
checkNodes(EE, [N|NN], SS, TT, RR) :-
	\+ member(N, SS), member(N, TT), \+ member(N, RR), !,
	findall(WI, (member(E, EE), E = [_, N, WI]), LWI),
	findall(WO, (member(E, EE), E = [N, _, WO]), LWO),
	findall(N, member(N, TT), LT), length(LT, WT),
	sum_list(LWI, WTI), sum_list(LWO, WTO),
	(   WTI =:= WT, WTO =< WT; eN2(N, WTI, WTO, WT), !, fail),
	checkNodes(EE, NN, SS, TT, RR).
% checks that receivers receive the correct amount of information
checkNodes(EE, [N|NN], SS, TT, RR) :-
	\+ member(N, SS), \+ member(N, TT), member(N, RR), !,
	findall(W, (member(E, EE), E = [_, N, W]), LW),
	findall(N, member(N, RR), LR), length(LR, WR),
	sum_list(LW, WT),
	(   WT =:= WR; eN3(N, WR, WT), !, fail),
	checkNodes(EE, NN, SS, TT, RR).
% checks source transceivers
checkNodes(EE, [N|NN], SS, TT, RR) :-
	member(N, SS), member(N, TT), \+ member(N, RR), !,
	findall(WI, (member(E, EE), E = [_, N, WI]), LWI),
	findall(WO, (member(E, EE), E = [N, _, WO]), LWO),
	findall(N, member(N, SS), LS), length(LS, WST),
	findall(N, member(N, TT), LT), length(LT, WTT),
	sum_list(LWI, WTI), sum_list(LWO, WTO),
	(   WTI =:= WTT; eN4a(N, WTI, WTT), !, fail),
	(   WTO - WTI =< WST; eN4b(N, WTO, WTI, WST), !, fail),
	checkNodes(EE, NN, SS, TT, RR).
% check receiver transceivers (star transceivers)
checkNodes(EE, [N|NN], SS, TT, RR) :-
	\+ member(N, SS), member(N, TT), member(N, RR), !,
	findall(WI, (member(E, EE), E = [_, N, WI]), LWI),
	findall(WO, (member(E, EE), E = [N, _, WO]), LWO),
	findall(N, member(N, TT), LT), length(LT, WTT),
	findall(N, member(N, RR), LR), length(LR, WRT),
	sum_list(LWO, WTO), sum_list(LWI, WTI),
	(   WTO =< WTT; eN5a(N, LWO, WTT), !, fail),
	(   WTI - WTO =:= WRT; eN5b(N, WTI, WTO, WRT), !, fail),
	checkNodes(EE, NN, SS, TT, RR).

checkNodes(_, _, _, _, _) :- ifail("node check missing case."), !, fail.

e1(X, E) :- pre, format("Node ~w in edge ~w is not in graph.~n", [X, E]).
e2(W, E) :- pre, format("Edge ~w cannot appear ~w in the solution (its width should be stricly positive).~n", [E, W]).
e3(E) :- pre, format("Edge ~w is not permitted because of collinear nodes.~n", [E]).
e4(E) :- pre, format("Edge ~w appears multiple time in the solution.", [E]).
e5(E) :- pre, format("Edge ~w appears in the solution in both directions.", [E]).
e6a(E) :- pre, format("Edge ~w starts in a receiver.", [E]).
e6b(E) :- pre, format("Edge ~w starts in a broadcast antenna.", [E]).
e6c(E) :- pre, format("Edge ~w enters into a source node.", [E]).

eN1(N, WT, WS) :- pre, format("Source ~w only offers ~w units of information, but there are ~w going out of it.~n ", [N, WS, WT]).
eN2(N, WTI, WTO, WT) :- pre, format("Transceiver ~w has width ~w, but it has ~w units of information incoming and ~w units of information outgoing.~n", [N, WT, WTI, WTO]).
eN3(N, WR, WT) :- pre, format("Receiver ~w has width ~w, but receives ~w units of information.~n", [N, WR, WT]).
eN4a(N, WTI, WTT) :- pre, format("Source transceiver ~w has width ~w but ~w incoming units of information.~n", [N, WTT, WTI]).
eN4b(N, WTO, WTI, WST) :- pre, format("Source transceiver ~w creates ~w units of information but it has an outgoing / incoming number of units of ~w / ~w. ~n", [N, WST, WTO, WTI]).
eN5a(N, WTO, WTT) :- pre, format("Star transceiver ~w has width ~w but ~w outgoing units of information.~n", [N, WTT, WTO]).
eN5b(N, WTO, WTI, WRT) :- pre, format("Star transceiver ~w should be left with ~w units of information but it has an incoming / outgoing number of units of ~w / ~w. ~n", [N, WRT, WTI, WTO]).

ifail(S) :- pre, format("internal failure, please report. Location: ~w ~n", [S]).



